#include <libsecret/secret.h>

// Qt
#include <QCoreApplication>
#include <QDebug>
#include <QLibrary>
#include <QThread>

#include "Core/helpers.h"
#include "Core/stash.h"

#include "libsecretkeystore.h"

namespace libsecret {
    const SecretSchema* encfsManagerSchema(void) {
        static const SecretSchema schema = {
            "com.moritzmolch.encfsmanager", SECRET_SCHEMA_NONE,
            {
                { "stash_directory", SECRET_SCHEMA_ATTRIBUTE_STRING }
            }
        };

        return &schema;
    }

    typedef gboolean (*secret_password_store_sync_t) (const SecretSchema *schema,
                                                    const gchar *collection,
                                                    const gchar *label,
                                                    const gchar *password,
                                                    GCancellable *cancellable,
                                                    GError **error,
                                                    ...);

    typedef gchar *(*secret_password_lookup_sync_t) (const SecretSchema *schema,
                                                     GCancellable *cancellable,
                                                     GError **error,
                                                     ...);

    typedef gboolean (*secret_password_clear_sync_t) (const SecretSchema *schema,
                                                    GCancellable *cancellable,
                                                    GError **error,
                                                    ...);


    typedef void (*secret_password_free_t) (gchar *password);

    typedef void (*g_error_free_t) (GError *error);

    static secret_password_store_sync_t secret_password_store_sync = nullptr;
    static secret_password_lookup_sync_t secret_password_lookup_sync = nullptr;
    static secret_password_clear_sync_t secret_password_clear_sync = nullptr;
    static secret_password_free_t secret_password_free = nullptr;
    static g_error_free_t g_error_free = nullptr;
}

bool LibSecretKeyStore::m_isInitialized = false;
bool LibSecretKeyStore::m_isAvailable = false;
QString LibSecretKeyStore::m_statusMessage = QLatin1String();


// ================================================================================================
LibSecretKeyStore::LibSecretKeyStore(QObject *parent) : KeyStore(parent)
{
    if (m_isInitialized) {
        return;
    }
    m_isInitialized = true;
    init();
}

void LibSecretKeyStore::retrievePasswordSync(QSharedPointer<Stash> stash)
{
    GError *error = NULL;

    gchar *password = libsecret::secret_password_lookup_sync(
            libsecret::encfsManagerSchema(),
            NULL,
            &error,
            "stash_directory", Helpers::pathToRelative(stash->stashDir()).toUtf8().data(),
            NULL);


    if (error == NULL) {
        QString s = QString::fromUtf8(password);
        libsecret::secret_password_free(password);
        if (!s.isEmpty()) {
            emit retrievePasswordFinished(true, s);
        } else {
            emit retrievePasswordFinished(false, tr("The password couldn't be found in the keyring."));
        }
    } else {
        emit retrievePasswordFinished(false, QString::fromUtf8(error->message));
        libsecret::g_error_free(error);
    }
}


bool LibSecretKeyStore::isAvailable()
{
    return init();
}

QString LibSecretKeyStore::statusMessage()
{
    return m_statusMessage;
}

void LibSecretKeyStore::storePassword(QSharedPointer<Stash> stash, const QString &password)
{
    if (!init()) {
        emit storePasswordFinished(false, statusMessage());
        return;
    }

    QThread *thread = QThread::create([this, stash, password]() {
        GError *error = NULL;

        libsecret::secret_password_store_sync(
                libsecret::encfsManagerSchema(),
                stash->keyringName().isEmpty() ? SECRET_COLLECTION_DEFAULT : stash->keyringName().toUtf8().data(),
                QString(QLatin1String("Encfs password for %1")).arg(Helpers::pathToRelative(stash->stashDir())).toUtf8().data(),
                password.toUtf8().data(),
                NULL,
                &error,
                "stash_directory", Helpers::pathToRelative(stash->stashDir()).toUtf8().data(),
                NULL);

        if (error == NULL) {
            emit _storePasswordFinished(true);
        } else {
            emit _storePasswordFinished(false, QString::fromUtf8(error->message));
            libsecret::g_error_free(error);
        }
    });
    connect(thread, &QThread::finished, thread, &QObject::deleteLater);
    connect(this, &LibSecretKeyStore::_storePasswordFinished, this, &LibSecretKeyStore::storePasswordFinished, Qt::QueuedConnection);
    thread->start();
}

void LibSecretKeyStore::retrievePassword(QSharedPointer<Stash> stash)
{
    if (!init()) {
        emit retrievePasswordFinished(false, statusMessage());
        return;
    }

    QThread *thread = QThread::create([this, stash]() {
        GError *error = NULL;

        gchar *password = libsecret::secret_password_lookup_sync(
                libsecret::encfsManagerSchema(),
                NULL,
                &error,
                "stash_directory", Helpers::pathToRelative(stash->stashDir()).toUtf8().data(),
                NULL);


        if (error == NULL) {
            QString s = QString::fromUtf8(password);
            if (!s.isEmpty()) {
                emit _retrievePasswordFinished(true, s);
            } else {
                emit _retrievePasswordFinished(false, tr("The password couldn't be found in the keyring."));
            }
            libsecret::secret_password_free(password);
        } else {
            emit _retrievePasswordFinished(false, QString::fromUtf8(error->message));
            libsecret::g_error_free(error);
        }
    });
    connect(thread, &QThread::finished, thread, &QObject::deleteLater);
    connect(this, &LibSecretKeyStore::_retrievePasswordFinished, this, &LibSecretKeyStore::retrievePasswordFinished, Qt::QueuedConnection);
    thread->start();
}

void LibSecretKeyStore::deletePassword(QSharedPointer<Stash> stash)
{
    if (!init()) {
        emit deletePasswordFinished(stash, statusMessage());
        return;
    }

    QThread *thread = QThread::create([this, stash]() {
        GError *error = NULL;

        libsecret::secret_password_clear_sync(
                libsecret::encfsManagerSchema(),
                NULL,
                &error,
                "stash_directory", Helpers::pathToRelative(stash->stashDir()).toUtf8().data(),
                NULL);

        if (error == NULL) {
            emit _deletePasswordFinished(true);
        } else {
            emit _deletePasswordFinished(false, QString::fromUtf8(error->message));
            libsecret::g_error_free(error);
        }
    });
    connect(thread, &QThread::finished, thread, &QObject::deleteLater);
    connect(this, &LibSecretKeyStore::_deletePasswordFinished, this, &LibSecretKeyStore::deletePasswordFinished, Qt::QueuedConnection);
    thread->start();
}

bool LibSecretKeyStore::init()
{
    if (m_isAvailable) {
        return true;
    }

    QLibrary lib_secret;
    for (int i = 0; i < 10; i++) {
        lib_secret.setFileName(QString(QLatin1String("libsecret-1.so.%1")).arg(QString::number(i)));
        if (lib_secret.load()) {
            break;
        }
    }

    if (!lib_secret.isLoaded()) {
        m_isAvailable = false;
        m_statusMessage = tr("Couldn't open libsecret-1.\n"
                             "Please make sure that it is installed and in the library path.");
        qDebug() << qPrintable(statusMessage());
        return false;
    }

    libsecret::secret_password_store_sync = (libsecret::secret_password_store_sync_t)lib_secret.resolve("secret_password_store_sync");
    libsecret::secret_password_lookup_sync = (libsecret::secret_password_lookup_sync_t)lib_secret.resolve("secret_password_lookup_sync");
    libsecret::secret_password_clear_sync = (libsecret::secret_password_clear_sync_t)lib_secret.resolve("secret_password_clear_sync");
    libsecret::secret_password_free = (libsecret::secret_password_free_t)lib_secret.resolve("secret_password_free");

    QLibrary lib_glib2(QLatin1String("glib-2.0"), 0);
    if (!lib_glib2.load()) {
        m_isAvailable = false;
        m_statusMessage = tr("Couldn't open libglib-2.0.\n"
                             "Please make sure that it is installed and in the library path.");
        qDebug() << qPrintable(statusMessage());
        return false;
    }

    libsecret::g_error_free = (libsecret::g_error_free_t)lib_glib2.resolve("g_error_free");

    m_isAvailable = true;
    return true;
}
