#include "mounter.h"

// Qt
#include <QApplication>
#include <QDebug>
#include <QScreen>

// Core
#include "Core/stashmanager.h"
#include "Core/stash.h"

#include "managerwindow.h"
#include "pager.h"
#include "pages/mountstashpage.h"
#include "pages/configurestashpage.h"
#include "pages/configurestashexpertpage.h"
#include "pages/stashinformationpage.h"
#include "pages/stashcommandpage.h"

Mounter *Mounter::m_instance = nullptr;

Mounter::Mounter(QObject *parent) : QObject(parent)
{
    if (m_instance != nullptr) {
        qDebug() << "Only one instance of Mounter is allowed";
    }
    m_instance = this;

    connect(&StashManager::instance(), &StashManager::mountFinished, [this](const QSharedPointer<Stash> &stash, bool success, const QString &message) {
        Q_UNUSED(message);

        if (success) {
            m_stashes.remove(stash);
        } else {
            mount(stash, message);
        }
    });
}

Mounter::~Mounter()
{
    m_instance = nullptr;
}

Mounter &Mounter::instance()
{
    return *m_instance;
}

void Mounter::mount(const QSharedPointer<Stash> &stash, const QString &message)
{
    if (!m_stashes.contains(stash)) {
        m_stashes.insert(stash, message);
    }

    if (m_mutex.tryLock()) {
        while (!m_stashes.isEmpty()) {
            const QSharedPointer<Stash> stash = m_stashes.firstKey();
            Pager pager(stash, ManagerWindow::instance());

            auto mountStashPage = new MountStashPage(stash);
            mountStashPage->mountResult(false, m_stashes.first());
            pager.addPage(mountStashPage);
            pager.addPage(new ConfigureStashPage(stash));
            pager.addPage(new ConfigureStashExpertPage(stash));
            pager.addPage(new StashCommandPage(stash));
            pager.addPage(new StashInformationPage(stash));
            pager.adjustSize();
            pager.selectPage(0);

            if (!ManagerWindow::instance()->isVisible()) {
                QSize screenSize = QApplication::screens().at(0)->size();
                pager.move(screenSize.width()/2 - pager.width()/2,
                     screenSize.height()/2 - pager.height()/2);
            }

            int result = pager.exec();
            if (result == QDialog::Rejected) {
                //m_stashes.clear();
                m_stashes.remove(stash);
            } else {
                m_stashes.remove(stash);
            }
            m_mutex.unlock();
        }
    }
}

void Mounter::cancel(const QSharedPointer<Stash> &stash)
{
    m_stashes.remove(stash);
}
