#include "addstashpage.h"
#include "ui_addstashpage.h"

// Qt
#include <QBoxLayout>
#include <QDebug>
#include <QDir>
#include <QFileDialog>
#include <QFileInfo>
#include <QIcon>
#include <QRegularExpression>

// Core
#include "Core/helpers.h"
#include "Core/settings.h"
#include "Core/stash.h"
#include "Core/stashmanager.h"

#include <MHelpBrowser>

AddStashPage::AddStashPage(const QSharedPointer<Stash> &stash, QWidget *parent) :
    QWizardPage(parent),
    ui(new Ui::AddStashPage),
    m_stash(stash),
    m_initialized(false)
{
    ui->setupUi(this);

    setPixmap(QWizard::LogoPixmap, QIcon::fromTheme(QLatin1String("list-add")).pixmap(24,24));

    ui->label_help->setText(QString(QLatin1String("<html><a href=\"stashadd.html\">%1</a></html>")).arg(tr("Help")));

    ui->lineEdit_stashDir->setMinimumHeight(ui->button_stashDir->height()-6);
    ui->lineEdit_mountDir->setMinimumHeight(ui->button_mountDir->height()-6);
    ui->lineEdit_password->setMinimumHeight(ui->button_mountDir->height()-6);
    ui->lineEdit_retypePassword->setMinimumHeight(ui->button_mountDir->height()-6);
}

AddStashPage::~AddStashPage()
{
    delete ui;
}

QString AddStashPage::stashDir() const
{
    if (ui->radioButton_stashDirDefault->isChecked()) {
        return ui->radioButton_stashDirDefault->text();
    } else {
        return ui->lineEdit_stashDir->text();
    }
}

QString AddStashPage::mountDir() const
{
    if (ui->radioButton_mountDirDefault->isChecked()) {
        return ui->radioButton_mountDirDefault->text();
    } else {
        return ui->lineEdit_mountDir->text();
    }
}

QString AddStashPage::password() const
{
    if (ui->lineEdit_password->text() == ui->lineEdit_retypePassword->text()) {
        return ui->lineEdit_password->text();
    } else {
        return QLatin1String();
    }
}

void AddStashPage::init()
{

}

void AddStashPage::setMessage(const QString &message)
{
    if (message.isEmpty()) {
        ui->label_message->hide();
    } else {
        ui->label_message->show();
        ui->label_message->setText(QStringLiteral("<span style=\"color:#ff0000;\">✘</span> ") + message);
    }
    adjustDialogSize();
}

void AddStashPage::adjustDialogSize()
{
    qobject_cast<QDialog*>(parent()->parent())->adjustSize();
    if (!Settings::instance().showSidebar()) {
        qobject_cast<QDialog*>(parent()->parent())->setMinimumHeight(sizeHint().height()+30);
        qobject_cast<QDialog*>(parent()->parent())->setMaximumHeight(sizeHint().height()+30);
    }
}

bool AddStashPage::validate()
{
    if (!m_stash->isAvailable()) {
        ui->line_password->show();
        ui->groupBox_password->show();
        ui->button_add->setText(tr("Create"));
    } else {
        ui->line_password->hide();
        ui->groupBox_password->hide();
        ui->button_add->setText(tr("Import"));
    }

    if (stashDir().isEmpty()) {
        setMessage(tr("Please provide a stash directory"));
        ui->button_add->setEnabled(false);
        adjustDialogSize();
        return false;
    }

    if (!StashManager::instance().stashByStashDir(m_stash->stashDir()).isNull()) {
        setMessage(tr("The stash is already configured"));
        ui->button_add->setEnabled(false);
        adjustDialogSize();
        return false;
    }

    if (mountDir().isEmpty()) {
        setMessage(tr("Please provide a mount directory"));
        ui->button_add->setEnabled(false);
        adjustDialogSize();
        return false;
    }

    if (!StashManager::instance().stashByMountDir(m_stash->mountDir()).isNull()) {
        setMessage(tr("The mount directory is already in use"));
        ui->button_add->setEnabled(false);
        adjustDialogSize();
        return false;
    }

    if (m_stash->stashDir() == m_stash->mountDir()) {
        setMessage(tr("The stash and mount directories are identical"));
        ui->button_add->setEnabled(false);
        adjustDialogSize();
        return false;
    }

    if (!m_stash->allowNonEmptyMountDirectory()) {
        QDir _mountDir(m_stash->mountDir());
        if (_mountDir.exists() && !_mountDir.isEmpty(QDir::NoDotAndDotDot)) {
            setMessage(tr("The mount directory is not empty"));
            ui->button_add->setEnabled(false);
            adjustDialogSize();
            return false;
        }
    }

    if (!m_stash->isAvailable()) {
        if (ui->lineEdit_password->text().isEmpty() && ui->lineEdit_retypePassword->text().isEmpty()) {
            setMessage(tr("Please enter a password"));
            ui->button_add->setEnabled(false);
            adjustDialogSize();
            return false;
        } else if (ui->lineEdit_password->text() != ui->lineEdit_retypePassword->text()) {
            setMessage(tr("Passwords do not match"));
            ui->button_add->setEnabled(false);
            adjustDialogSize();
            return false;
        }
    }

    setMessage(QStringLiteral());
    ui->button_add->setEnabled(true);
    adjustDialogSize();
    return true;
}

void AddStashPage::generateDefaultStashDir()
{
    QString defaultStashDirBase = Helpers::pathToAbsolute(QString(QLatin1String(".%1")).arg(tr("Private")));
    QString defaultStashDir = defaultStashDirBase;
    for (int i = 1; ; i++) {
        if ( i > 1) {
            defaultStashDir = defaultStashDirBase + QString(QLatin1String(" (%1)")).arg(QString::number(i));
        }

        if (StashManager::instance().stashByStashDir(defaultStashDir).isNull()) {
            break;
        }
    }

    ui->radioButton_stashDirDefault->setText(defaultStashDir);
}

void AddStashPage::generateDefaultMountDir()
{
    QString proposedMountDir = QFileInfo(stashDir()).fileName();
    proposedMountDir.remove(QRegularExpression(QLatin1String("^\\.|-encfs$|_encfs$")));

    if (proposedMountDir.isEmpty()) {
        proposedMountDir = Settings::instance().baseDir().replace(QLatin1String("$HOME"), QDir::homePath());
    }

    ui->radioButton_mountDirDefault->setText(Helpers::pathToAbsolute(proposedMountDir));
    if (ui->radioButton_mountDirDefault->isChecked()) {
        m_stash->setMountDir(ui->radioButton_mountDirDefault->text());
    }
}

void AddStashPage::on_label_help_linkActivated(const QString &link)
{
    MHelpBrowser::open(link);
}

void AddStashPage::on_button_cancel_clicked()
{
    if (m_stash->isBusy() || ui->button_add->isChecked()) {
        StashManager::instance().cancel(m_stash);
    } else {
        qobject_cast<QDialog*>(parent()->parent())->reject();
    }
}

void AddStashPage::on_button_add_clicked()
{
    if (m_stash->isAvailable()) {
        m_stash->setIsInitialized(true);
        StashManager::instance().save();
        qobject_cast<QDialog*>(parent()->parent())->accept();
    } else {
        StashManager::instance().create(m_stash, ui->lineEdit_password->text());
    }
}


void AddStashPage::setVisible(bool visible)
{
    if (m_initialized) {
        QWizardPage::setVisible(visible);
        return;
    }
    m_initialized = true;

    generateDefaultStashDir();
    generateDefaultMountDir();

    connect(m_stash.data(), &Stash::isBusyChanged, ui->button_add, &MSpinnerPushButton::setDisabled);
    connect(m_stash.data(), &Stash::isBusyChanged, ui->button_add, &MSpinnerPushButton::setChecked);

    m_stash->setStashDir(ui->radioButton_stashDirDefault->text());
    connect(ui->radioButton_stashDirDefault, &QRadioButton::toggled, ui->lineEdit_stashDir, &QLineEdit::setDisabled);
    connect(ui->radioButton_stashDirDefault, &QRadioButton::toggled, ui->button_stashDir, &QPushButton::setDisabled);
    connect(ui->radioButton_stashDirDefault, &QRadioButton::toggled, [this](bool checked) {
        if (checked) {
            m_stash->setStashDir(ui->radioButton_stashDirDefault->text());
        } else {
            m_stash->setStashDir(ui->lineEdit_stashDir->text());
        }
    });

    m_stash->setMountDir(ui->radioButton_mountDirDefault->text());
    connect(ui->radioButton_mountDirDefault, &QRadioButton::toggled, ui->lineEdit_mountDir, &QLineEdit::setDisabled);
    connect(ui->radioButton_mountDirDefault, &QRadioButton::toggled, ui->button_mountDir, &QPushButton::setDisabled);
    connect(ui->radioButton_mountDirDefault, &QRadioButton::toggled, [this](bool checked) {
        if (checked) {
            m_stash->setMountDir(ui->radioButton_mountDirDefault->text());
        } else {
            m_stash->setMountDir(ui->lineEdit_mountDir->text());
        }
    });

    ui->button_sidebar->setChecked(Settings::instance().showSidebar());
    connect(ui->button_sidebar, &QPushButton::toggled, &Settings::instance(), &Settings::setShowSidebar);
    connect(&Settings::instance(), &Settings::showSidebarChanged, ui->button_sidebar, &QPushButton::setChecked);

    //ui->lineEdit_stashDir->setText(m_stash->stashDir());
    connect(m_stash.data(), &Stash::stashDirChanged, this, &AddStashPage::stashDirChanged);
    connect(ui->lineEdit_stashDir, &QLineEdit::textChanged, m_stash.data(), &Stash::setStashDir);
    connect(ui->lineEdit_stashDir, &QLineEdit::textChanged, this, &AddStashPage::generateDefaultMountDir);
    connect(ui->radioButton_stashDirDefault, &QRadioButton::toggled, this, &AddStashPage::generateDefaultMountDir);

    //ui->lineEdit_mountDir->setText(m_stash->mountDir());
    connect(m_stash.data(), &Stash::mountDirChanged, this, &AddStashPage::mountDirChanged);
    connect(ui->lineEdit_mountDir, &QLineEdit::textChanged, m_stash.data(), &Stash::setMountDir);


    connect(ui->radioButton_stashDirDefault, &QRadioButton::toggled, this, &AddStashPage::validate);
    connect(ui->radioButton_mountDirDefault, &QRadioButton::toggled, this, &AddStashPage::validate);

    connect(ui->lineEdit_stashDir, &QLineEdit::textChanged, this, &AddStashPage::validate);
    connect(ui->lineEdit_mountDir, &QLineEdit::textChanged, this, &AddStashPage::validate);
    connect(ui->lineEdit_password, &QLineEdit::textChanged, this, &AddStashPage::validate);
    connect(ui->lineEdit_retypePassword, &QLineEdit::textChanged, this, &AddStashPage::validate);

    connect(m_stash.data(), &Stash::isAvailableChanged, this, &AddStashPage::validate);

    connect(m_stash.data(), &Stash::createFinished, this, &AddStashPage::createResult);

    connect(m_stash.data(), &Stash::isBusyChanged, this, &AddStashPage::stashIsBusyChanged);

    validate();
    QWizardPage::setVisible(visible);
}

void AddStashPage::createResult(bool success, const QString &message)
{
    if (success) {
        qobject_cast<QDialog*>(parent()->parent())->accept();
    } else {
        setMessage(message);
    }
}

void AddStashPage::stashIsBusyChanged(bool busy)
{
    ui->radioButton_stashDirDefault->setDisabled(busy);
    ui->radioButton_stashDirCustom->setDisabled(busy);
    ui->lineEdit_stashDir->setDisabled(busy);

    ui->radioButton_mountDirDefault->setDisabled(busy);
    ui->radioButton_mountDirCustom->setDisabled(busy);
    ui->lineEdit_mountDir->setDisabled(busy);

    ui->lineEdit_password->setDisabled(busy);
    ui->lineEdit_retypePassword->setDisabled(busy);

    if (busy) {
        ui->label_message->clear();
        adjustDialogSize();
    }

    ui->button_add->setDisabled(busy);
}

void AddStashPage::stashDirChanged(const QString &stashDir)
{
    if (ui->radioButton_stashDirDefault->isChecked()) {
        return;
    }

    if (ui->lineEdit_stashDir->text() == stashDir) {
        return;
    }

    ui->lineEdit_stashDir->setText(stashDir);
}

void AddStashPage::mountDirChanged(const QString &mountDir)
{
    if (ui->radioButton_mountDirDefault->isChecked()) {
        return;
    }

    if (ui->lineEdit_mountDir->text() == mountDir) {
        return;
    }

    ui->lineEdit_mountDir->setText(mountDir);
}

void AddStashPage::on_lineEdit_stashDir_returnPressed()
{
    ui->button_add->click();
}

void AddStashPage::on_lineEdit_mountDir_returnPressed()
{
    ui->button_add->click();
}

void AddStashPage::on_lineEdit_password_returnPressed()
{
    ui->button_add->click();
}

void AddStashPage::on_lineEdit_retypePassword_returnPressed()
{
    ui->button_add->click();
}

void AddStashPage::on_button_stashDir_clicked()
{
    QFileDialog fileDialog(this);
    fileDialog.setDirectory(ui->lineEdit_stashDir->text().isEmpty() ? Helpers::pathToAbsolute(Settings::instance().baseDir()) : ui->lineEdit_stashDir->text());
    fileDialog.setFileMode(QFileDialog::Directory);
    fileDialog.setOptions(QFileDialog::ShowDirsOnly);

    if(fileDialog.exec()) {
       ui->lineEdit_stashDir->setText(fileDialog.selectedFiles().first());
    }
}

void AddStashPage::on_button_mountDir_clicked()
{
    QFileDialog fileDialog(this);
    fileDialog.setDirectory(ui->lineEdit_mountDir->text().isEmpty() ? Helpers::pathToAbsolute(Settings::instance().baseDir()) : ui->lineEdit_mountDir->text());
    fileDialog.setFileMode(QFileDialog::Directory);
    fileDialog.setOptions(QFileDialog::ShowDirsOnly);

    if(fileDialog.exec()) {
       ui->lineEdit_mountDir->setText(fileDialog.selectedFiles().first());
    }
}
