#include "stashtableview.h"
#include "ui_stashtableview.h"

// Qt
#include <QAbstractTableModel>
#include <QApplication>
#include <QDebug>
#include <QtWidgets>

// Core
#include <Core/stash.h>
#include <Core/stashmanager.h>

#include "checkboxitemdelegate.h"
#include "managerwindow.h"
#include "selectdisableditemdelegate.h"
#include "stashmimedata.h"


/*! This is used to draw the dropindicator through the whole row instead just over / under the cell
 *  Mostly taken from: https://stackoverflow.com/questions/7596584/qtreeview-draw-drop-indicator
 */
class StashTableView::DropIndicatorProxyStyle : public QProxyStyle {
public:
    DropIndicatorProxyStyle(QStyle* style = nullptr) : QProxyStyle(style) {}

    void drawPrimitive ( PrimitiveElement element, const QStyleOption * option, QPainter * painter, const QWidget * widget = 0 ) const {
        if (element == QStyle::PE_IndicatorItemViewItemDrop && !option->rect.isNull()){
            QStyleOption opt(*option);
            opt.rect.setLeft(0);
            if (widget) opt.rect.setRight(widget->width());
            //QProxyStyle::drawPrimitive(element, &opt, painter, widget);
            painter->fillRect(opt.rect.adjusted(0,0,0,1), QColor(0,0,0,100));
            return;
        }
        QProxyStyle::drawPrimitive(element, option, painter, widget);
    }
};


class StashTableView::Model : public QAbstractTableModel
{
    Q_OBJECT

public:
    explicit Model(StashTableView *parent = nullptr) :
        QAbstractTableModel(parent),
        m_stashGroup()
    {
        connect(parent, &StashTableView::stashGroupChanged, [this]() {
            beginResetModel();
            endResetModel();
        });
        connect(&StashManager::instance(), &StashManager::isMountedChanged, [this](const QSharedPointer<Stash> &stash, bool isMounted) {
            Q_UNUSED(isMounted)
            int row = StashManager::instance().indexOf(m_stashGroup, stash);
            if (row == -1) { return; }
            emit dataChanged(createIndex(row, 0), createIndex(row, 2)); // redraw whole row to redraw the light blue focusing indicator
        });
        connect(&StashManager::instance(), &StashManager::isAvailableChanged, [this](const QSharedPointer<Stash> &stash, bool isAvailable) {
            Q_UNUSED(isAvailable)
            int row = StashManager::instance().indexOf(m_stashGroup, stash);
            if (row == -1) { return; }
            emit dataChanged(createIndex(row, 0), createIndex(row, 2));
        });
        connect(&StashManager::instance(), &StashManager::mountDirChanged, [this](const QSharedPointer<Stash> &stash, const QString &mountDir) {
            Q_UNUSED(mountDir)
            int row = StashManager::instance().indexOf(m_stashGroup, stash);
            if (row == -1) { return; }
            emit dataChanged(createIndex(row, 0), createIndex(row, 2));
        });
        connect(&StashManager::instance(), &StashManager::stashDirChanged, [this](const QSharedPointer<Stash> &stash, const QString &stashDir) {
            Q_UNUSED(stashDir)
            int row = StashManager::instance().indexOf(m_stashGroup, stash);
            if (row == -1) { return; }
            emit dataChanged(createIndex(row, 0), createIndex(row, 2));
        });

        connect(&StashManager::instance(), &StashManager::stashRemoved, [this](const QSharedPointer<Stash> &stash) {
            Q_UNUSED(stash)
            beginResetModel();
            endResetModel();
        });

        connect(&StashManager::instance(), &StashManager::stashAdded, [this](const QSharedPointer<Stash> stash) {
            Q_UNUSED(stash)
            beginResetModel();
            endResetModel();
        });

        connect(&StashManager::instance(), &StashManager::orderChanged, [this]() {
            beginResetModel();
            endResetModel();
        });

        connect(&StashManager::instance(), &StashManager::stashGroupChanged, [this]() {
            beginResetModel();
            endResetModel();

        });
    }

    ~Model() {}

    int rowCount(const QModelIndex &) const override { return StashManager::instance().size(m_stashGroup); }
    int columnCount(const QModelIndex &) const override { return 3; }
    virtual Qt::DropActions supportedDropActions() const override { return Qt::MoveAction; }
    virtual Qt::DropActions supportedDragActions() const override { return Qt::MoveAction; }

    virtual QStringList mimeTypes() const override
    {
        return QStringList() << "application/vnd.moritzmolch.encfsmanager.stash";
    }

    QVariant headerData(int section, Qt::Orientation, int role) const override
    {
        if (role == Qt::DisplayRole) {
            switch (section) {
            case 0:
                //~ comment "The Mount Directory"
                return tr("Mount Directory");
            case 1:
                //~ comment "The Stash Directory"
                return tr("Stash Directory");
            case 2:
                //~ comment the heading over the checkboxes in the stash overview
                return tr("Mounted");
            }
        }
        return QVariant();
    }

    Qt::ItemFlags flags(const QModelIndex &index) const override
    {
        Qt::ItemFlags flags = Qt::ItemNeverHasChildren | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsEnabled | Qt::ItemIsEditable;

        if (index.isValid()) {
            if ((StashManager::instance().size(m_stashGroup) > index.row()) && StashManager::instance().at(m_stashGroup, index.row())->isAvailable())
            {
                flags |= Qt::ItemIsUserCheckable;
            }

            flags |= Qt::ItemIsDragEnabled;
        } else {
            flags |= Qt::ItemIsDropEnabled;
        }

        return flags;
    }

    QVariant data(const QModelIndex &index, int role) const override
    {
        switch (role) {
        case Qt::DisplayRole:
            switch (index.column()) {
            case 0:
                return StashManager::instance().at(m_stashGroup, index.row())->mountDir();
                break;
            case 1:
                return StashManager::instance().at(m_stashGroup, index.row())->stashDir();
                break;
            }
            break;
        case Qt::CheckStateRole:
            switch (index.column()) {
            case 2:
                return StashManager::instance().at(m_stashGroup, index.row())->isMounted();
                break;
            }
        }
        return QVariant();
    }

    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        if (indexes.size() >= 0 && indexes.at(0).isValid() && indexes.at(0).row() >= 0 && indexes.at(0).row() < StashManager::instance().size(m_stashGroup)) {
            StashMimeData *mimeData = new StashMimeData;
            mimeData->setStash(StashManager::instance().at(m_stashGroup, indexes.at(0).row()));
            return mimeData;
        }
        return new QMimeData();
    }

    virtual bool dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) override
    {
        Q_UNUSED(data) Q_UNUSED(action) Q_UNUSED(column) Q_UNUSED(parent)
        if (row < 0) {
            return false;
        }

        StashTableView *view = qobject_cast<StashTableView*>(this->parent());
        QSharedPointer<Stash> selectedStash = view->selectedStash();
        int draggedRow = StashManager::instance().indexOf(m_stashGroup, selectedStash);

        if (row > draggedRow) {
            row--;
        }

        StashManager::instance().move(m_stashGroup, draggedRow, row);
        view->ui->tableView->selectRow(row);

        return true;
    }

    QVariant m_stashGroup;
};

/*
/// The model: use applyFilter() to filter stashgroups =============================================
class StashTableViewModel : public QAbstractTableModel
{
    Q_OBJECT

public:
    explicit StashTableViewModel(StashTableView *view, QObject *parent = nullptr);
    ~StashTableViewModel() {}

    int rowCount(const QModelIndex &) const override { return m_stashes.size(); }
    int columnCount(const QModelIndex &) const override { return 3; }
    QVariant headerData(int section, Qt::Orientation, int role) const override {
        if (role == Qt::DisplayRole) { return m_headerTitles.at(section); } return QVariant();
    }
    QVariant data(const QModelIndex &index, int role) const override;
    Qt::ItemFlags flags(const QModelIndex &index) const override;
    virtual Qt::DropActions supportedDropActions() const override { return Qt::MoveAction; }
    virtual Qt::DropActions supportedDragActions() const override { return Qt::MoveAction; }
    virtual bool dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) override;

    QVariant stashGroup() const { return m_stashGroup; }

public slots:
    void setStashGroup(const QVariant &stashGroup = QVariant()) {
        beginResetModel();
        m_stashes = StashManager::instance().stashes(stashGroup);
        m_stashGroup = stashGroup;
        endResetModel();
    }

private:
    QStringList m_headerTitles;
    StashTableView *m_view;
    QVariant m_stashGroup;
    QList<QSharedPointer<Stash>> m_stashes;

    // QAbstractItemModel interface
public:
    virtual QStringList mimeTypes() const override
    {
        return QStringList() << "application/vnd.moritzmolch.encfsmanager.stash";
    }

    // QAbstractItemModel interface
public:
    virtual QMimeData *mimeData(const QModelIndexList &indexes) const override;
};

QMimeData *StashTableViewModel::mimeData(const QModelIndexList &indexes) const
{
    if (indexes.size() >= 0 && indexes.at(0).isValid() && indexes.at(0).row() >= 0 && indexes.at(0).row() < m_stashes.size()) {
        StashMimeData *mimeData = new StashMimeData;
        mimeData->setStash(m_stashes.at(indexes.at(0).row()));
        return mimeData;
    }
    return new QMimeData();
}

Qt::ItemFlags StashTableViewModel::flags(const QModelIndex &index) const
{
    Qt::ItemFlags flags = Qt::ItemNeverHasChildren | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsEnabled | Qt::ItemIsEditable;

    if (index.isValid()) {
        if ((m_stashes.size() > index.row()) && m_stashes.at(index.row())->isAvailable())
        {
            flags |= Qt::ItemIsUserCheckable;
        }

        flags |= Qt::ItemIsDragEnabled;
    } else {
        flags |= Qt::ItemIsDropEnabled;
    }

    return flags;
}

QVariant StashTableViewModel::data(const QModelIndex &index, int role) const
{
    switch (role) {
    case Qt::DisplayRole:
        switch (index.column()) {
        case 0:
            return m_stashes.at(index.row())->mountDir();
            break;
        case 1:
            return m_stashes.at(index.row())->stashDir();
            break;
        }
        break;
    case Qt::CheckStateRole:
        switch (index.column()) {
        case 2:
            return m_stashes.at(index.row())->isMounted();
            break;
        }
    }
    return QVariant();
}

bool StashTableViewModel::dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent)
{
    bool moveToLastRow = row >= m_stashes.size();

    if ((row < 0) || (row >= m_stashes.size())) {
        row = m_stashes.size()-1;
    }

    // Map filtered list onto the complete list
    int draggedRow = StashManager::instance().indexOf(m_view->draggedStash());
    int newRow = StashManager::instance().indexOf(m_stashes.at(row));

    if ((draggedRow < newRow) && (newRow > 0)) {
        newRow--;
    }

    if (newRow >= StashManager::instance().size()) {
        newRow = StashManager::instance().size()-1;
    }

    if (moveToLastRow) {
        newRow++;
    }

    StashManager::instance().move(draggedRow, newRow);

    applyFilter(filter());

    if (m_stashes.contains(m_view->draggedStash())) {
        m_view->selectRow(m_stashes.indexOf(m_view->draggedStash()));
    }
    return true;
}*/

/*StashTableViewModel::StashTableViewModel(StashTableView *view, QObject *parent) :
    QAbstractTableModel(parent),
    m_headerTitles(QStringList() << tr("Mount Directory") << tr("Stash Directory") << tr("Mounted")),
    m_view(view)
{
    setStashGroup();
    connect(&StashManager::instance(), &StashManager::isMountedChanged, [this](const QSharedPointer<Stash> stash, bool isMounted) {
        int row = m_stashes.indexOf(stash);
        if (row == -1) { return; }
        emit dataChanged(createIndex(row, 0), createIndex(row, 2));
    });
    connect(&StashManager::instance(), &StashManager::isAvailableChanged, [this](const QSharedPointer<Stash> stash, bool isAvailable) {
        int row = m_stashes.indexOf(stash);
        if (row == -1) { return; }
        emit dataChanged(createIndex(row, 0), createIndex(row, 2));
    });
    connect(&StashManager::instance(), &StashManager::mountDirChanged, [this](const QSharedPointer<Stash> stash, const QString &mountDir) {
        int row = m_stashes.indexOf(stash);
        if (row == -1) { return; }
        emit dataChanged(createIndex(row, 0), createIndex(row, 2));
    });

    connect(&StashManager::instance(), &StashManager::stashRemoved, [this](int i) {
        setStashGroup(m_stashGroup);
    });

    connect(&StashManager::instance(), &StashManager::stashAdded, [this](const QSharedPointer<Stash> stash) {
        setStashGroup(m_stashGroup);
    });

    connect(&StashManager::instance(), &StashManager::orderChanged, [this]() {
        setStashGroup(m_stashGroup);
    });

    connect(&StashManager::instance(), &StashManager::stashGroupChanged, [this]() {
        setStashGroup(m_stashGroup);

    });
}*/


/// StashTableView Implementation =================================================================
StashTableView::StashTableView(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::StashTableView)
{
    ui->setupUi(this);

    m_model = new Model(this);

    CheckBoxItemDelegate *checkboxDelegate = new CheckBoxItemDelegate(this);
    connect(checkboxDelegate, &CheckBoxItemDelegate::toggled, [this](QModelIndex index, bool state) {
        const QSharedPointer<Stash> stash = StashManager::instance().at(m_model->m_stashGroup, index.row());
        if (stash->isBusy()) {
            stash->cancel();
        } else {
            if (state) {
                StashManager::instance().mount(stash);
            } else {
                StashManager::instance().unmount(stash);
            }
        }
    });

    ui->tableView->setItemDelegateForColumn(2, checkboxDelegate);
    ui->tableView->setItemDelegateForColumn(1, new SelectDisabledItemDelegate(ui->tableView));
    ui->tableView->setItemDelegateForColumn(0, new SelectDisabledItemDelegate(ui->tableView));

    ui->tableView->setModel(m_model);
    ui->tableView->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
    ui->tableView->horizontalHeader()->setSectionResizeMode(2, QHeaderView::ResizeToContents);
    ui->tableView->horizontalHeader()->setMinimumHeight(24);
    ui->tableView->verticalHeader()->setSectionResizeMode(QHeaderView::Fixed);

    ui->tableView->setStyle(new DropIndicatorProxyStyle(style()));

//    connect(this, &StashTableView::stashGroupChanged, m_model, &StashTableViewModel::setStashGroup);

    connect(ui->tableView->model(), &QAbstractItemModel::modelReset, [this]() {
        emit selectedStashChanged(selectedStash());
    });

    connect(ui->tableView->selectionModel(), &QItemSelectionModel::selectionChanged, [this]() {
        emit selectedStashChanged(selectedStash());
    });

    /*connect(ui->tableView, &QTableView::pressed, [this]() {
        setDraggedRow(ui->tableView->selectionModel()->selectedRows().at(0).row());
    });*/

    connect(ui->tableView, &QTableView::doubleClicked, [&]() {
        ManagerWindow::instance()->on_actionConfigureStash_triggered();
    });
}

StashTableView::~StashTableView()
{
    delete ui;
}

QSharedPointer<Stash> StashTableView::selectedStash() const
{
    if (ui->tableView->selectionModel()->selectedRows().size() > 0) {
        int row = ui->tableView->selectionModel()->selectedRows().at(0).row();
        if ((row >= 0) && (row < StashManager::instance().size(stashGroup()))) {
            return StashManager::instance().stashes(stashGroup()).at(row);
        }
    }
    return QSharedPointer<Stash>();
}

QVariant StashTableView::stashGroup() const { return m_model->m_stashGroup; }
void StashTableView::setStashGroup(const QVariant &stashGroup)
{
    if (m_model->m_stashGroup == stashGroup) {
        return;
    }

    m_model->m_stashGroup = stashGroup;
    emit stashGroupChanged(stashGroup);
}


#include "stashtableview.moc"
