#include "trayicon.h"

//Qt
#include <QApplication>
#include <QDebug>
#include <QMap>
#include <QMenu>
#include <QSystemTrayIcon>

//Core
#include "Core/settings.h"
#include <Core/stashmanager.h>
#include <Core/stash.h>

#include "managerwindow.h"

TrayIcon *TrayIcon::m_instance = nullptr;

void TrayIcon::updateIcon()
{
    if (!m_systemTrayIcon.isVisible()) {
        return;
    }

    if (StashManager::instance().isBusy()) {
        m_systemTrayIcon.setIcon(m_iconBusy);
    } else {
        bool indicateMounted = false;
        foreach (auto stash, StashManager::instance().stashes()) {
            if (stash->trayIconIndicateMounted() && stash->isMounted()) {
                indicateMounted = true;
                break;
            }
        }

        if (indicateMounted) {
            m_systemTrayIcon.setIcon(m_iconMounted);
        } else {
            m_systemTrayIcon.setIcon(m_iconIdle);
        }
    }
}

TrayIcon::TrayIcon()
{
    if (m_instance != nullptr) {
        qDebug() << "Warning: Only one instance of TrayIcon is allowed";
        return;
    }
    m_instance = this;

    m_iconIdle = QIcon(QLatin1String(":/icons/encfsmanager/48/gnome-encfs-manager-panel.png"));
    m_iconBusy = QIcon(QLatin1String(":/icons/encfsmanager/48/gnome-encfs-manager-panel-attention.png"));
    m_iconMounted = QIcon(QLatin1String(":/icons/encfsmanager/48/gnome-encfs-manager-panel-mounted.png"));

    m_systemTrayIcon.setIcon(m_iconIdle);
    m_systemTrayIcon.setContextMenu(new QMenu());

    QObject::connect(&StashManager::instance(), &StashManager::isBusyChanged, [this]() {
        updateIcon();
    });

    QObject::connect(&StashManager::instance(), &StashManager::trayIconIndicateMountedChanged, [this]() {
        updateIcon();
    });

    QObject::connect(&StashManager::instance(), &StashManager::showInTrayChanged, [this]() {
        update();
    });

    QObject::connect(&StashManager::instance(), &StashManager::shortNameChanged, [this]() {
        update();
    });

    QObject::connect(&StashManager::instance(), &StashManager::stashGroupChanged, [this]() {
        update();
    });

    QObject::connect(&StashManager::instance(), &StashManager::isMountedChanged, [this]() {
        update();
        updateIcon();
    });

    QObject::connect(&StashManager::instance(), &StashManager::isAvailableChanged, [this]() {
        update();
        updateIcon();
    });

    QObject::connect(&StashManager::instance(), &StashManager::orderChanged, [this]() {
        update();
    });

    QObject::connect(&StashManager::instance(), &StashManager::stashAdded, [this]() {
        update();
        updateIcon();
    });

    QObject::connect(&StashManager::instance(), &StashManager::stashRemoved, [this]() {
        update();
        updateIcon();
    });

    QObject::connect(&Settings::instance(), &Settings::trayIconEnabledChanged, [this](bool trayIconEnabled) {
        m_systemTrayIcon.setVisible(trayIconEnabled);
        if (trayIconEnabled) {
            updateIcon();
            update();
        }
    });

    QObject::connect(&Settings::instance(), &Settings::trayIconStashGroupsChanged, [this]() {
        update();
    });

    QObject::connect(&Settings::instance(), &Settings::stashGroupsChanged, [this]() {
        update();
    });

    m_systemTrayIcon.setVisible(Settings::instance().trayIconEnabled());
    updateIcon();
    update();
}

TrayIcon::~TrayIcon()
{
    delete(m_systemTrayIcon.contextMenu());
    m_instance = nullptr;
}

void TrayIcon::update()
{
    if (m_instance == nullptr) {
        return;
    }

    if (!m_instance->m_systemTrayIcon.isVisible()) {
        return;
    }

    QMenu *menu = m_instance->m_systemTrayIcon.contextMenu();
    menu->clear();

    QAction *action_header = menu->addAction(QObject::tr("Encrypted folders:"));
    action_header->setEnabled(false);

    if (Settings::instance().trayIconStashGroups()) {
        QMap<QString, QMenu*> stashgroups;
        foreach (auto stash, StashManager::instance().stashes()) {
            if (stash->isAvailable() && stash->showInTray()) {
                QString stash_stashGroup = stash->stashGroup();
                if (!stash_stashGroup.isEmpty() && !stashgroups.contains(stash_stashGroup)) {
                    stashgroups.insert(stash_stashGroup, new QMenu(stash_stashGroup, menu));
                }
                QAction *action;
                if (stash_stashGroup.isEmpty()) {
                    action = menu->addAction(stash->mountDir().splitRef('/').last().toString());
                } else {
                    action = stashgroups.value(stash_stashGroup)->addAction(stash->mountDir().splitRef('/').last().toString());
                }
                action->setCheckable(true);
                action->setChecked(stash->isMounted());
                QObject::connect(action, &QAction::triggered, [stash](bool checked) {
                    if(checked) {
                        update();
                        StashManager::instance().mount(stash);
                    } else {
                        update();
                        StashManager::instance().unmount(stash);
                    }
                });
            }
        }

        if (!stashgroups.isEmpty()) {
            menu->addSeparator();
        }

        foreach (const QString &stashGroup, Settings::instance().stashGroups()) {
            if (stashgroups.contains(stashGroup)) {
                menu->addMenu(stashgroups.value(stashGroup));
            }
        }

    } else {
        foreach (auto stash, StashManager::instance().stashes()) {
            if (stash->isAvailable() && stash->showInTray()) {
                QAction *action = menu->addAction(stash->mountDir().splitRef('/').last().toString());
                action->setCheckable(true);
                action->setChecked(stash->isMounted());
                QObject::connect(action, &QAction::triggered, [stash](bool checked) {
                    if(checked) {
                        StashManager::instance().mount(stash);
                    } else {
                        StashManager::instance().unmount(stash);
                    }
                });
            }
        }
    }


    // Unmount All Menu entry =====================================================================
    menu->addSeparator();
    QAction *action_unmountAll = menu->addAction(QObject::tr("Unmount all"), []() {
        foreach (auto stash, StashManager::instance().stashes()) {
            if (stash->isMounted()) {
                StashManager::instance().unmount(stash);
            }
        }
    });
    action_unmountAll->setEnabled(false);
    foreach (auto stash, StashManager::instance().stashes()) {
        if (stash->isMounted()) {
            action_unmountAll->setEnabled(true);
            break;
        }
    }


    // Show Manager Menu entry ====================================================================
    menu->addSeparator();
    bool managerWindowVisible = (ManagerWindow::instance() != nullptr) && (ManagerWindow::instance()->isVisible());
    QAction *action_showManager = menu->addAction(QObject::tr("Show Manager"), []() {
        bool managerWindowVisible = (ManagerWindow::instance() != nullptr) && (ManagerWindow::instance()->isVisible());
        if (managerWindowVisible) {
            //delete(ManagerWindow::instance());
            ManagerWindow::instance()->setVisible(false);
        } else {
            if (ManagerWindow::instance() == nullptr) {
                new ManagerWindow;
            }
            ManagerWindow::instance()->setVisible(true);
        }
    });
    action_showManager->setCheckable(true);
    action_showManager->setChecked(managerWindowVisible);


    // Quit Menu entry ============================================================================
    menu->addSeparator();
    menu->addAction(QObject::tr("Quit"), []() {
        StashManager::instance().shutdown();
    });

    //QMenu *oldMenu = m_instance->m_systemTrayIcon.contextMenu();
    //m_instance->m_systemTrayIcon.setContextMenu(menu);
    //delete(oldMenu);
}

TrayIcon &TrayIcon::instance()
{
    if (m_instance == nullptr) {
        new TrayIcon;
    }
    return *m_instance;
}

bool TrayIcon::isAvailable()
{
    /** QSystemTrayIcon::isSystemTrayAvailable() isn't reliable
     * with the appindicator implementation in gnome shell,
     * we assume there's a tray */
    //return QSystemTrayIcon::isSystemTrayAvailable();
    return true;
}
